#
# python t4view_v3.py [option] predefReport
#    where
#         predefReport: module containing the PredefinedReports dictionary
#

import cgitb
import cgi
import urllib
import sys
import gdchart
import MySQLdb
import Cheetah.Template
import wasd
import pyXLWriter
import datetime
from optparse import OptionParser
import mx.DateTime
import mx.DateTime.ISO

try:
    import rdb
    import types
    class Cursor:
        def __init__(self):
            self._curs = None
        def _check_executed(self):
            if not self._curs:
                raise "execute() first"
        def execute(self, stmt, args = None):
            """Execute a query.

            query -- string, query to execute on server
            args -- optional sequence or mapping, parameters to use with query.
            """
            stmt = stmt.replace('%s','?')
            self._curs = rdb.statement(stmt)
            if args is None:
                self._curs.execute()
            elif type(args) in (types.TupleType, types.ListType):
                self._curs.execute(*args)
            else:
                self._curs.execute(args)
            return None
        def fetchall(self):
            """Fetchs all available rows from the cursor."""
            self._check_executed()
            return self._curs.fetchall()
except:
    pass

global OutputMain, OutputResult, Backend, PredefinedReports
Tables_name = None
Items_name = None
Items_info = None

UseJPEG = False

def split_isots(isots):
    date,time = isots.split(' ')
    year, month, day = date.split('-')
    hour, minute = time.split(':')
    return int(year), int(month), int(day), int(hour), int(minute)

def escape_string(string):
    "Remove any offending from a string used in a MySQL command"
    return string.replace(';', '').replace("'", '')

def normalize_table_name(table_name):
    "Remove ':' and replace '.' by '_' into table_name"
    return table_name.replace('.', '_').replace(':', '')

def extract_names(name):
    """
    return
            - None if Sample Time column
        or
            - table name, column name
    """
    name = name.lstrip()
    fclass, fname = name.split(']')
    if fname == 'Sample Time': # new table
        return None
    else:
        return normalize_table_name(fclass.replace('[','')), name

def get_tables(schema, curs):
    "Return all table names except table name node and repository"
    global Tables_name
    if Tables_name is None:
        req = "select distinct table_name from %s.repository" % (schema,)
        curs.execute(req)
        res = curs.fetchall()
        Tables_name = [v[0].rstrip() for v in res]
    return Tables_name

def get_node_id(schema, node_name, curs):
    "Convert a node name to a node id"
    global Backend
    req = "select node_id from %s.node " % (schema,)
    req += "where node_name = %s"
    if Backend == 'mysql':
        node_name = escape_string(node_name)
    curs.execute(req, node_name)
    return curs.fetchall()[0][0]

def get_timestamp_range(schema, node_name, curs):
    "Return min and max timestamp for a node using first table found table"
    global Backend
    node_name = escape_string(node_name)
    node_id = get_node_id(schema, node_name, curs)
    tbl_name = get_tables(schema, curs)[0]
    if Backend == 'mysql':
        stmt = "select date_format(min(sample_time),'%%Y-%%m-%%d %%H:%%i:%%s'),"
        stmt += "date_format(max(sample_time),'%%Y-%%m-%%d %%H:%%i:%%s')"
        stmt += "from %s.%s" %(schema, tbl_name) + " where node_id = %s"
        curs.execute(stmt, (node_id,))
        return curs.fetchall()[0]
    elif Backend == 'rdb':
        stmt = "select min(sample_time), max(sample_time)"
        stmt += " from " + tbl_name + " where node_id = ?"
        curs = rdb.statement(stmt)
        curs.execute(node_id)
        return curs.fetchall()[0]

def get_items_name_id(schema, curs):
    "Return all item names"
    global Items_name
    req = "select item_id, item_name from %s.repository order by table_name, item_id" % (schema,)
    if Items_name is None:
        curs.execute(req)
        res = curs.fetchall()
        Items_name = [(v[0], v[1].rstrip()) for v in res]
    return Items_name

def get_item_infos(schema, curs):
    "Return associate table name and column name for all items"
    global Items_info
    if Items_info is None:
        req = "select item_id, table_name, column_name, item_name "
        req += "from %s.repository order by item_id" % (schema,)
        curs.execute(req)
        res = curs.fetchall()
        rep = {}
        for (id, table, col, item) in res:
            rep[id] = (table.rstrip(), col.rstrip(), item.rstrip())
        Items_info = rep
    return Items_info

def get_node_names(schema, curs):
    "Return all node names"
    global Backend
    req = "select node_name from %s.node" % (schema,)
    curs.execute(req)
    res = curs.fetchall()
    return [v[0].rstrip() for v in res]

def get_values(schema, schema_data, node_name, item_info, start, end, curs):
    "Return values for a requested item"
    global Backend
    # remove SQL offending characters from needed string
    node_name = escape_string(node_name)
    start = escape_string(start)
    end = escape_string(end)
    node_id = get_node_id(schema, node_name, curs)
    tbl = item_info[0]
    colname = item_info[1]
    # fetch sample_time without second part and data value
    # for the requested node interval
    if Backend == 'mysql':
        stmt = "select date_format(sample_time,'%Y-%m-%d %H:%i'),"
        stmt += '"%s" ' % (colname,)
        stmt += 'from %s.%s where node_id = %s and ' % (schema_data, tbl, node_id)
        stmt += "sample_time >= '%s' and sample_time <= '%s'" % (start, end)
        curs.execute(stmt)
        res = curs.fetchall()
        return [v[0] for v in res], [v[1] for v in res]
    elif Backend == 'rdb':
        stmt = "select sample_time,"
        stmt += '"%s" ' % (colname,)
        stmt += 'from %s.%s where node_id = ? and ' % (schema_data, tbl,)
        stmt += "sample_time >= ? and sample_time <= ?"
        curs = rdb.statement(stmt)
        curs.execute(node_id, start, end)    
        res = curs.fetchall()
        return [v[0].strftime("%Y-%m-%d %H:%M") for v in res], [v[1] for v in res]

def colnames_cvt(lst, names):
    lst = lst[:]
    # convert form column number to integer, invalid value are replaced by 0
    for i in range(len(lst)):
        try:
            lst[i] = int(lst[i])
            names[lst[i]]
        except:
            # replace invalid key by a valid one
            lst[i] = names.iterkeys().next()
    # convert column numbers to column names
    return lst, [names[i] for i in lst]
    

def execmain(form, schema, curs):
    "Display the main form"
    global PredefinedReports, OutputResult
    query = None
    start = None
    end = None
    print 'Content-type: text/html\n\n'
    if form.has_key('start'):
        start = cgi.escape(str(form.getfirst('start')))
    if form.has_key('end'):
        end = cgi.escape(str(form.getfirst('end')))
    node_names = get_node_names(schema, curs)
    item_infos = get_items_name_id(schema, curs)
    ts = get_timestamp_range(schema, node_names[0], curs)

    OutputMain.db = schema
    OutputMain.node_names = node_names
    OutputMain.item_infos = item_infos
    OutputMain.start = start or str(ts[0])[:19]
    OutputMain.end = end or str(ts[1])[:19]
    OutputMain.predef_reports = PredefinedReports
    print OutputMain

def execresult(form, schema, curs):
    global PredefinedReports, OutputResult, UseJPEG
    # red, blue, green, yellow, orange, ...
    colors = (0xff0000, 0x3232cc, 0x25b415, 0xffff15, 0xff7f00,
              0xff00ff, 0x00ffff, 0xffffff, 0xffffff)
    # 
    start = cgi.escape(str(form.getfirst('start')))
    end = cgi.escape(str(form.getfirst('end')))
    node_name = cgi.escape(str(form.getfirst('node')))
    resolution = cgi.escape(str(form.getfirst('resolution'))).lower()
    chart_type = cgi.escape(str(form.getfirst('chart_type')))
    html_or_chart = form.getfirst('html_or_chart')
    form_colnames = form.getlist('col')
    form_colnames = form_colnames[0:7]
    predef_report = cgi.escape(str(form.getfirst('predef')))

    try:
        mx_start = mx.DateTime.ISO.ParseDateTime(start)
        mx_end = mx.DateTime.ISO.ParseDateTime(end)
    except:
        execmain(form, schema, curs)
        return

    if mx_end <= mx_start:
        execmain(form, schema, curs)
        return

    if resolution == 'hour':
        schema_data = "%s_h" % (schema,)
        if Backend == 'mysql':
            curs.execute("use %s_h" % (schema,))
        elif Backend == 'rdb':
            curs.execute("set schema '%s_h'" % (schema,))
    else: #resolution == 'minute'
        # maximum range is one day
        if mx_start + 1 < mx_end:
            execmain(form, schema, curs)
            return
        schema_data = schema
        if Backend == 'mysql':
            curs.execute("use %s" % (schema,))
        elif Backend == 'rdb':
            curs.execute("set schema '%s'" % (schema,))

    # collect all columns infos for all tables
    item_infos = get_item_infos(schema, curs)

    if html_or_chart == 'HTML' and chart_type == 'spreadsheet':
        if predef_report == 'None':
            form_colnames, colnames = colnames_cvt(form_colnames, item_infos)
            report = [[chart_type] + form_colnames]
        else:
            report = PredefinedReports[predef_report]

        print 'Content-Type: application/vnd.ms-excel'
        download = True
        if download:
            # print 'Content-Disposition: attachment'
            pass
        else:
            print 'Content-Disposition: inline'
            # print 'Content-Length: %s' % len(buffer)
        print
        sys.stdout.flush()

        xls_colprefix = 'BCDEFGHIJKLMNOPQRSTUVWXYZ'
        libel = []
        item_requested = []
        for v in report:
            # convert column numbers to column names
            libel += [item_infos[i][2] for i in v[1:]]
            item_requested+= [item_infos[i] for i in v[1:]]

        wb = pyXLWriter.Writer("sys$output")
        ws = wb.add_worksheet()
        
        ws.write("A1", 'Sample Time')
        for i in range(len(libel)):
            ws.write(xls_colprefix[i] + '1', libel[i])
        
        ts_range, values = get_values(schema, schema_data, node_name,
                                      item_requested[0], start, end, curs)
        for j in range(len(ts_range)):
            ts = split_isots(ts_range[j])
            ws.write('A%d' % (j + 2,), datetime.datetime(*ts))
            ws.write('B%d' % (j + 2,), values[j])

        for i in range(1, len(libel)):
            ts_range, values = get_values(schema, schema_data, node_name,
                                          item_requested[i], start, end, curs)
            for j in range(len(values)):
                ws.write('%s%d' % (xls_colprefix[i], j + 2), values[j])
        wb.close()
        return

    if html_or_chart == 'HTML':
        OutputResult.node_name = node_name
        OutputResult.start = start
        OutputResult.end = end

        if predef_report == 'None':
            form_colnames, colnames = colnames_cvt(form_colnames, item_infos)
            report = [[chart_type] + form_colnames]
        else:
            report = PredefinedReports[predef_report]

        lst_imgsrc = []
        for v in report:
            imgsrc = '?chart_type=' + v[0]
            imgsrc += '&node=' + node_name
            imgsrc += '&resolution=' + resolution
            imgsrc += '&start=' + urllib.quote_plus(start)
            imgsrc += '&end=' + urllib.quote_plus(end)
            imgsrc += '&html_or_chart=CHART'
            for colname in v[1:]:
                imgsrc += '&col=%d' % colname
            lst_imgsrc.append(imgsrc)
        OutputResult.imgsrc = lst_imgsrc

        lst_libel = []
        for v in report:
            # convert column numbers to column names
            colnames = [item_infos[i][2] for i in v[1:]]
            for i in range(9 - len(colnames)):
                colnames.append('')
            libel = [[], [], []]
            lst_libel.append(libel)
            for i in range(len(colnames)):
                if (i < 3):
                    libel[0].append(["%06X" % colors[i], colnames[i]])
                elif (i < 6):
                    libel[1].append(["%06X" % colors[i], colnames[i]])
                else:
                    libel[2].append(["%06X" % colors[i], colnames[i]])

        OutputResult.libel = lst_libel
        print 'Content-type: text/html\n\n'
        print OutputResult
    else: # html_or_chart != 'HTML'
        form_colnames, items = colnames_cvt(form_colnames, item_infos)
        all_values = []
        ts_range = []
        for item in items:
#            colname = colname.replace('&lt;', '<')
            ts_range, values = get_values(schema, schema_data, node_name,
                                          item, start, end, curs)
            all_values.append(values)
        if chart_type == 'lines':
            sgraph = gdchart.Line()
        else: # stacked areas
            sgraph = gdchart.Area()
            # "SUM" for stack_type is buggy
            # sgraph.stack_type = "SUM"
            for i in range(len(all_values) - 1):
                for j in range(len(all_values[0])):
                    all_values[i + 1][j] += all_values[i][j]
        sgraph.width = 800
        sgraph.height = 450
        # sgraph.xtitle = ""
        # sgraph.ytitle = "Value"
        if len(items) == 1:
            sgraph.title = items[0][2]
        sgraph.setData(*all_values)
        sgraph.setLabels(ts_range)
        sgraph.bg_color = 0xcccccc
        sgraph.border = ("ALL",)
        sgraph.grid = "LABELS"
        sgraph.set_color = colors
        if UseJPEG:
            sgraph.image_type = "JPEG"
            print 'Content-Type: image/jpeg\n'
        else:
            print 'Content-Type: image/png\n'
        sgraph.draw(sys.stdout)

def dispatch(schema, curs):
    "Main routine"
    global Backend
    i = 0
    while 1:
        print "Script-Control: X-stream-mode"
        request_method = wasd.cgi_info('REQUEST_METHOD')

        form = cgi.FieldStorage()
        
        if Backend == 'rdb':
            rdb.statement("set transaction read only").execute()
        # if all necessary fields are present display the HTML/PNG|JPEG
        # results page
        # else display the main page
        if (form.getfirst('start') and form.getfirst('end') and
            form.getfirst('node') and form.getfirst('resolution') and
            form.getfirst('chart_type') and
            form.getfirst('html_or_chart') and
            (len(form.getlist('col')) > 0 or
             form.getfirst('predef') != 'None')):
            execresult(form, schema, curs)
        else:
            execmain(form, schema, curs)
        if Backend == 'rdb':
            rdb.statement("commit").execute()

        if not wasd.isCgiPlus(): break
        wasd.cgi_eof()
        i += 1
        if (i > 100):
            sys.exit()
            # block until next request 
        wasd.cgi_info("")

if __name__ == '__main__':
    usage = "usage: python %prog [options] database|rdb_schema predefReport"
    parser = OptionParser(usage = usage)
    parser.add_option("-b", "--backend", action="store", type="choice",
                      choices=["rdb", "mysql"], default="mysql",
                      help="Choice storage backend")
    # mysql specific
    parser.add_option("-n", "--node", action="store", type="string",
                      dest="host", default="localhost",
                      help="MySQL, connect to host")
    parser.add_option("-u", "--user", action="store", type="string",
                      dest="user", default="",
                      help="MySQL, user for login if not current user")
    parser.add_option("-p", "--password", action="store", type="string",
                      dest="password", default="",
                      help="MySQL, password to use when connecting to server")
    parser.add_option("-d", "--database", action="store", type="string",
                      dest="schema", help="name of mysql database to used")
    # rdb specific
    parser.add_option("-r", "--rdb_database", action="store", type="string",
                      dest="rdb_db", default="SQL$DATABASE",
                      help="Rdb, database used")
    parser.add_option("-s", "--schema", action="store", type="string",
                      dest="schema", help="name of rdb schema to used")
    (options, args) = parser.parse_args()

    Backend = options.backend.lower()
    PredefinedReports = __import__(args[0]).PredefinedReports

    try:
        parser = OptionParser(usage = usage)
        host = options.host
        user = options.user
        pwd = options.password
        schema = options.schema.upper()
        wasd.init()
    except:
        sys.exit()

    cgitb.enable(display=1, logdir='../python_tmp')

    templateError = 't4_error.tmpl'

    if Backend == 'mysql':
        try:
            # Connect to the MySQL database
            cnx = MySQLdb.connect(host=host, db=schema, user=user, passwd=pwd)
            curs = cnx.cursor()
            # curs.execute('set autocommit = 0')
            curs.execute("SET SESSION sql_mode='ansi'")
        except:
            output = Cheetah.Template.Template(file=templateError)
            output.reason = "MySQL connection failure"
            print 'Content-type: text/html\n\n'
            print output
            sys.exit()
    elif Backend == 'rdb':
        try:
            # Connect to the Rdb database
            rdb.statement("attach 'filename %s'" % (options.rdb_db,)).execute()
            rdb.statement("set schema '%s'" % (options.schema,)).execute()
            curs = Cursor()
        except:
            output = Cheetah.Template.Template(file=templateError)
            output.reason = "Rdb connection failure"
            print 'Content-type: text/html\n\n'
            print output
            sys.exit()

    templateMain = 't4_view_v3.tmpl'
    templateResult = 't4_result_v3.tmpl'

    OutputMain = Cheetah.Template.Template(file=templateMain)
    OutputResult = Cheetah.Template.Template(file=templateResult)

    try:
        dispatch(schema, curs)
    except:
        output = Cheetah.Template.Template(file=templateError)
        output.reason = "unknown failure"
        print 'Content-type: text/html\n\n'
        print output
